'' This is the support library for the Propeller Meta Compiler.
'' It's based on the Meta2 Compiler-Compiler written originally
'' for the IBM 360 by Michael Green and that, in turn, is based
'' on the Meta Compiler-Compiler presented at a 1963 ACM
'' conference on programming languages.

OBJ
   mass : "BB_massStorage"       ' EEPROM and SD card interface
   ser  : "BB_FullDuplexSerial"  ' Serial I/O routines

CON
   maxSymbol = 128               ' Maximum symbol length
   maxBackup = 128               ' Maximum saved characters
   maxOutput = 128               ' EEPROM page size
   maxFileNm = 12                ' Filename length
   maxLine   = 80                ' Maximum line length for debugging

   bspChar   = $08               ' ASCII backspace
   tabChar   = $09               ' ASCII tab
   lfChar    = $0A               ' ASCII line feed
   crChar    = $0D               ' ASCII carriage return
   quote     = 34                ' ASCII double quote

VAR
   long inputAddr                ' Input EEPROM address
   long outputAddr               ' Output EEPROM address
   long value                    ' Holds scanned numeric value
   long indent                   ' Number of spaces to indent
   long debug                    ' True if in debug mode
   long lineOriented             ' True if CR / LF is significant
   long commentType              ' "S" for Spin, "C" for C
   byte fifoInPtr
   byte fifoOutPtr               ' Input FIFO
   byte fifoBuffer[maxBackup]
   byte symBuf[maxSymbol+1]      ' Holds scanned symbol
   byte outputPtr
   byte outputBuf[maxOutput+1]   ' Holds EEPROM output page
   byte debugInPtr
   byte debugInBuf[maxLine+1]    ' Holds current input line for debugging
   byte debugOutPtr
   byte debugOutBuf[maxLine+1]   ' Holds current output line for debugging

PUB start(s,d) | c               ' Initialize support library
   debug := d
   ser.start(mass#bootRx,mass#bootTx,%0000,9600)
   ser.str(string("Meta2 Library v1.005",crChar,lfChar))
   if \mass.start
      ser.str(string("Can't start I2C/SPI",crChar,lfChar))
      return false
   fifoInPtr := fifoOutPtr := 0
   debugInPtr := debugOutPtr := 0
   lineOriented := false
   commentType := "S"
   indent := 0
   if s <> 0                     ' File name supplied?
      if \mass.mountSDVol(mass#spiDO,mass#spiClk,mass#spiDI,mass#spiCS)
         ser.str(string("Can't mount SD card",crChar,lfChar))
         return false
      if s == -1
         askFilename(string("Input"))
         if symBuf[0] <> 0       ' Use EEPROM if no file name
            ifnot copyToEEPROM(@symBuf)
               ser.str(string("No such file",crChar,lfChar))
               return false
      else
         ifnot copyToEEPROM(s)   ' Use supplied file name
            ser.str(string("Can't open '"))
            ser.str(s)
            ser.str(string("'",crChar,lfChar))
            return false
      mass.unmountSDVol
   inputAddr := mass#bootAddr + $10000 ' Source input at $10000
   outputAddr := mass#bootAddr + $18000 ' Output at $18000
   outputPtr := 0
   ser.str(string("Beginning of compilation",crChar,lfChar))
   return true

PRI copyToEEPROM(s) | c          ' Copy file to EEPROM
   if \mass.openFile(s,"r") <> 0
      return false               ' Can't open the file
   outputAddr := mass#bootAddr + $10000 ' Write to input
   repeat
      c := \mass.readFile(@outputBuf,maxOutput)
      if c == -1                 ' No more blocks to read
         quit
      if \mass.writeEEPROM(outputAddr,@outputBuf,c)
         abort string("copyToEEPROM - writeEEPROM error")
      if \mass.writeWait(outputAddr)
         abort string("copyToEEPROM - writeWait error")
      if c < maxOutput           ' Partial block read
         outputAddr += c
         quit
      outputAddr += maxOutput    ' On to next block
   c := 0                        ' Now write zero terminator
   if \mass.writeEEPROM(outputAddr,@c,1)
      abort string("copyToEEPROM - writeEEPROM EOF error")
   if \mass.writeWait(outputAddr)
      abort string("copyToEEPROM - writeWait EOF error")
   return true

PUB stop(s) | c                  ' Finish up processing
   ser.str(string("End of compilation",crChar,lfChar))
   outputBuf[outputPtr++] := 0
   if \mass.writeEEPROM(outputAddr,@outputBuf,outputPtr)
      ser.str(string("writeEEPROM error",crChar,lfChar))
      return false
   if \mass.writeWait(outputAddr)
      ser.str(string("writeWait error",crChar,lfChar))
      return false
   if s <> 0                     ' File name supplied?
      if \mass.mountSDVol(mass#spiDO,mass#spiClk,mass#spiDI,mass#spiCS)
         ser.str(string("Can't mount SD card",crChar,lfChar))
         return false
      if s == -1
         askFilename(string("Output"))
         if symBuf[0] <> 0       ' Use EEPROM if no file name
            ifnot copyFromEEPROM(@symBuf)
               return false
      else
         ifnot copyFromEEPROM(s)
            return false
      mass.unmountSDVol
   return true

PRI copyFromEEPROM(s)            ' Copy EEPROM to file
   if \mass.openFile(s,"w") <> 0
      ser.str(string("Can't create file",crChar,lfChar))
      return false
   inputAddr := mass#bootAddr + $18000 ' Read from output
   outputBuf[maxOutput]~         ' Force zero byte terminator
   repeat
      if \mass.readEEPROM(inputAddr,@outputBuf,maxOutput)
         ser.str(string("readEEPROM error",crChar,lfChar))
         return false
      if outputPtr := strsize(@outputBuf)
         inputAddr += outputPtr
         if \mass.writeFile(@outputBuf,outputPtr) < 0
            ser.str(string("writeFile error",crChar,lfChar))
            return false
   until outputPtr < maxOutput
   \mass.closeFile               ' All done
   ser.str(string("Output copied to file",crChar,lfChar))
   return true

PRI askFilename(s) | i           ' Get file name into symBuf
   ser.str(s)
   ser.str(string(" file? "))
   ser.rxflush
   i := 0
   repeat
      case symBuf[i] := ser.rx
         bspChar:                ' Backspace
            if i
               ser.str(string(bspChar," ",bspChar))
               i--
         crChar:                 ' Carriage return
            symBuf[i]~           ' Zero byte terminator
            ser.str(string(crChar,lfChar))
            quit
         other:                  ' Other characters,
            if i < maxFileNm     '  save up to maxFilenm
               ser.tx(symBuf[i++])

PUB checkAbort(a) | c            ' Check for error exit
   if a == true
      return true                ' Successful completion
   if a == false
      a := string("Syntax error")
   ser.str(a)
   ser.str(string(crChar,lfChar))
   debugInBuf[debugInPtr]~       ' Zero terminator
   ser.str(string(">>"))
   ser.str(@debugInBuf)          ' Display debug buffer
   ser.str(string("<<",crChar,lfChar))
   return false

PUB exit
   waitcnt(clkfreq/10 + cnt)     ' Wait for any serial output to finish
   reboot

PUB message(s)                   ' Copy the string to the console
   ser.str(s)

PUB messageNewline               ' Copy a newline to the console
   ser.str(string(crChar,lfChar))

PRI newChar                      ' Return the next input character
   if fifoInPtr == fifoOutPtr
      if \mass.readEEPROM(inputAddr,@result,1)
         abort string("newChar - readEEPROM error")
      if result                  ' Zero byte marks end of input
         inputAddr++             ' If not end, advance pointer
      if (inputAddr & $3FFFF) > $17FFF
         abort string("newChar - end of EEPROM")
      if result and result <> lfChar
         if result == crChar     ' Save input line for debugging
            debugInBuf[debugInPtr]~
            debugInPtr := 0      ' Prepare for next input line
            if debug
               ser.str(string(">>"))
               ser.str(@debugInBuf)
               ser.str(string("<<",crChar,lfChar))
               waitcnt(clkfreq/20 + cnt) ' Wait a bit for the terminal
         else
            if debugInPtr < maxLine
               debugInBuf[debugInPtr++] := result
      fifoBuffer[fifoInPtr++] := result
      fifoInPtr //= maxBackup    ' Save character in FIFO
      fifoOutPtr := fifoInPtr
   else
      result := fifoBuffer[fifoOutPtr++]
      fifoOutPtr //= maxBackup   ' Get next character out of FIFO

PRI backupChar(n)                ' Back up n characters in input buffer
   fifoOutPtr := (fifoOutPtr + maxBackup - n) // maxBackup

PRI skipWhite | c, x, n          ' Skip white space
   repeat
      case c := newChar
         " ", tabChar:
            next
         lfChar, crChar:         ' If line oriented, CR/LF not whitespace
            if lineOriented
               backupChar(1)
               quit
         "/":                    ' Handle C type comments with /* ... */
            if commentType <> "C"
               backupChar(1)
               quit
            if newChar <> "*"
               backupChar(2)
               quit
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  "*":
                     if newChar == "/"
                        quit
         "'":                    ' Handle quote to end of line comments
            if commentType <> "S"
               backupChar(1)
               quit
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  lfChar, crChar:
                     backupChar(1)
                     quit
         "{":                    ' Handle nested braces comments
            if commentType <> "S"
               backupChar(1)
               quit
            n := 1
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  "{":
                     n++         ' Count nesting level
                  "}":
                     ifnot --n   ' Still nested, continue skipping
                        quit
         other:
            backupChar(1)        ' Back up one character
            quit       

PUB getValue                     ' Return the numeric value just scanned in
   return value

PUB setValue(v)                  ' Set the value variable
   value := v

PUB getIndent                    ' Return column number of first non-whitespace
   skipWhite                     '  after a span of whitespace.  May not work
   return debugInPtr             '   if called when currently not in whitespace

PUB setLineOriented(b)           ' Set lineOriented
   lineOriented := b

PUB setCommentType(t)            ' Set commentType
   commentType := t

PUB getEndline | c               ' Match CR or LF or CR/LF if lineOriented
   skipWhite
   ifnot lineOriented
      return false               ' Never matches if not lineOriented
   case newChar
      crChar:
         if newChar <> lfChar
            backupChar(1)
         return true
      lfChar:
         return true
      other:
         backupChar(1)
         return false
   

PUB getLiteral(target) | i, s    ' Match against supplied string
   skipWhite
   s := strsize(target)
   if s > maxBackup
      abort string("literal - maxBackup size")
   repeat i from 0 to s-1
      if byte[target][i] <> newChar
         backupChar(i+1)
         return false
   return true

PUB getIdent | i, c              ' Does .id
   skipWhite
   case c := newChar
      "_", "A".."Z", "a".."z":   ' Initial letter or "_"
         i := 0
         symBuf[i++] := c        ' Save character
      other:
         backupChar(1)
         return false            ' Not identifier
   repeat
      case c := newChar
         "_", "0".."9", "A".."Z", "a".."z":
            if i < maxSymbol     ' Save 1st part of identifier
               symBuf[i++] := c
            next                 ' Look for more
         other:
            symBuf[i]~           ' Zero byte terminator
            backupChar(1)
            return true

PRI checkNumber(base,digit)      ' Do value := value * base + digit
   if value ** base <> 0         ' Check for pending 32 bit overflow
      abort string("checkNumber - overflow base")
   value *= base
   if (value + digit) & $8000_0000 == 0 and value & $8000_0000 <> 0
      abort string("checkNumber - overflow digit")
   return value := value + digit

PUB getNumber | c, digits        ' Does .number
   skipWhite
   value := 0
   digits := 0
   case c := newChar
      "%":                       ' Binary or quaternary value
         if (c := newChar) == "%"
            repeat               ' Quaternary value
               case c := newChar
                  "_":
                     next
                  "0".."3":
                     checkNumber(4,c-"0")
                     digits++
                     next
                  other:
                     backupChar(1)
                     return digits > 0
         else
            fifoOutPtr := (fifoInPtr - 1) // maxBackup
            repeat               ' Binary value
               case c := newChar
                  "_":
                     next
                  "0".."1":
                     checkNumber(2,c-"0")
                     digits++
                     next
                  other:
                     backupChar(1)
                     return digits > 0
      "$":                       ' Hexadecimal value
         repeat
            case c := newChar
               "_":
                  next
               "0".."9":
                  checkNumber(16,c-"0")
                  digits++
                  next
               "A".."F":
                  checkNumber(16,c-"A"+10)
                  digits++
                  next
               "a".."f":
                  checkNumber(16,c-"a"+10)
                  digits++
                  next
               other:
                  backupChar(1)
                  return digits > 0
      "0".."9":
         checkNumber(10,c-"0")
         digits++
      other:
         backupChar(1)
         return digits > 0
   repeat
      case c := newChar
         "_":
            next
         "0".."9":
            checkNumber(10,c-"0")
            digits++
            next
         other:
            backupChar(1)
            return digits > 0

PRI storeString(c,p)
   symBuf[p++] := c
   if p == maxOutput
      abort string("string - too long")

PUB getString | p, c, s          ' Does .string
   skipWhite
   if newChar <> quote           ' Open quote?
      backupChar(1)
      return false
   p := 0
   s := 1                        ' State = Just started
   repeat
      case c := newChar
         lfChar, crChar:         ' End of line in string
            abort string("string - end of line")
         quote:
            if newChar == quote  ' Escaped quote
               if s == 2         ' End a substring with a quote
                  storeString(quote,p++)
               if s <> 1         ' Not at the start of the string
                  storeString(",",p++)
               storeString("3",p++)
               storeString("4",p++)
               s := 3            ' Quote placed in string
            else
               backupChar(1)
               if s == 2         ' End a substring with a quote
                 storeString(quote,p++)
               symBuf[p]~        ' Zero byte terminator
               return true
         other:
            if c < " "           ' Invalid character
               abort string("string - invalid")
            if s == 3            ' Previous character was a quote
               storeString(",",p++)
            if s <> 2            ' Start new substring with quote
               storeString(quote,p++)
            storeString(c,p++)
            s := 2               ' At least one non-quote stored

PUB getContent | p, c            ' Like .string, but string content only
   skipWhite
   if (c := newChar) <> quote    ' Open quote?
      backupChar(1)
      return false
   p := 0
   repeat
      case c := newChar
         lfChar, crChar:         ' End of line in string
            abort string("string - end of line")
         quote:
            if newChar == quote ' Escaped quote
               storeString(c,p++)
            else
               backupChar(1)
               symBuf[p]~        ' Zero byte terminator
               return true
         other:
            if c < " "           ' Invalid character
               abort string("string - invalid")
            storeString(c,p++)

PUB convertHex(digits) | v, p    ' Convert value to hexadecimal string
  p := 0
  symBuf[p++] := "$"             ' Leading "$"
  v := value << (8 - digits) << 2
  repeat digits
    symBuf[p++] := lookupz((v <-= 4) & $F : "0".."9", "A".."F")
  symBuf[p]~                     ' Zero byte terminator

PUB convertDec | i, p, x, v      ' Convert value to decimal string
   p := x := 0
   if (v := value) < 0           ' Leading sign if negative
      -v
      symBuf[p++] := "-"
   i := 1_000_000_000
   repeat 10                     ' Up to 10 digits
      if v => i
         symBuf[p++] := v / i + "0"
         v //= i
         x~~                     ' Always have at least 1 digit
      elseif x or i == 1
         symBuf[p++] := "0"
      i /= 10
   symBuf[p]~                    ' Zero byte terminator

DAT
spaces    byte  "                "  ' A total of 32 spaces
          byte  "                "
          byte  0

PUB outputIndent                 ' Indent # of spaces requested
   output(@spaces+(32-indent))

PUB adjustIndent(adj)            ' Adjust # of spaces to indent
   indent += adj
   if indent < 0 or indent > 32
      abort string("adjustIndent - out of range")

PUB outputNewline                ' Output a carriage return / line feed
   if debug
      debugOutBuf[debugOutPtr]~  ' Zero termination byte
      ser.str(string("|"))
      ser.str(@debugOutBuf)      ' Show saved output line
      ser.str(string("|",crChar,lfChar))
      waitcnt(clkfreq/20 + cnt)  ' Wait a bit for the terminal to catch up
   output(string(crChar,lfChar))
   debugOutPtr := 0

PUB output(s) | c                ' Copy string to output
   if s == 0
      s := @symBuf               ' Use symbol buffer if null
   repeat while c := byte[s++]
      if debugOutPtr < maxLine   ' Save output line for debugging
         debugOutBuf[debugOutPtr++] := c
      outputBuf[outputPtr++] := c
      if outputPtr == maxOutput
         if \mass.writeEEPROM(outputAddr,@outputBuf,maxOutput)
            abort string("output - writeEEPROM error")
         if \mass.writeWait(outputAddr)
            abort string("output - writeWait error")
         outputAddr += maxOutput
         if (outputAddr & $3FFFF) > $1FFFF
            abort string("output - end of EEPROM")
         outputPtr := 0

PUB key                          ' Return entered key (0 is an allowed value)
   return ser.rx

PUB keytime(ms) : c              ' Return entered key (within ms timeout)
   if (c := ser.rxtime(ms)) < 0  '  or return 0 if no key entered in time
      c~

PUB keycheck : c                 ' Return entered key or 0 if no key present
   if (c := ser.rxcheck) < 0
      c~

PUB str(s)                       ' Display supplied string
   ser.str(s)

PUB dec(v)                       ' Display signed decimal number
   ser.dec(v)

PUB hex(v,d)                     ' Display hexadecimal number of given width
   ser.hex(v,d)

PUB bin(v,d)                     ' Display binary number of given width
   ser.bin(v,d)

PUB out(c)                       ' Display given character
   ser.tx(c)